''' The simplest interpreter you can imagine. Currently there are 4 commands and 0 queries

    Other ideas to implement:
        - query for value of variable as a string
        - execute arbitrary line of code, return its return value as a string
        - execute a macro
'''
from __future__ import print_function
from lygadgets import message, isGSI
import lyipc.server
import os
import traceback

if not isGSI():
    raise RuntimeError('Non-klayout serving does not make sense')
import pya


def quiet_load_layout(filename, mode=0):
    ''' Swallow the error if the file is not completely written yet,
        or if the last file was not completely rendered.
        This doesn't seem to fatally affect the program.

        Modes are
        - 0 (default): replacing the current layout view
        - 1: making a new view
        - 2: adding the layout to the current view (mode 2)
    '''
    main = pya.Application.instance().main_window()
    try:
        view = main.load_layout(filename, mode)
    except RuntimeError as err:
        if err.args[0].split()[0] in ['Stream', 'Unexpected']:
            print(err)
        else:
            raise


def parse_message(msg):
    ''' Takes a msg read from the socket and does something with it.

        Careful of name conflict with lygadgets.

        The returned payload is not the same as what is returned from the called function:
        It is prepended with a status token and encoded (as a str) to be sent back over the socket
    '''
    return_val = None
    tokens = msg.split(' ')
    try:
        # if msg == 'kill':
        #     message('Stopping server -- remote shutdown')
        #     pya.Application.exit(pya.Application.instance())

        if tokens == ['reload', 'view']:
            main = pya.Application.instance().main_window()
            main.cm_reload()

        elif tokens[0] == 'load':
            filename = os.path.realpath(tokens[1])
            message(filename)
            if len(tokens) > 2:
                mode = int(tokens[2])
                quiet_load_layout(filename, mode)
            else:
                quiet_load_layout(filename)

        else:
            message('Received {}'.format(msg))

    except Exception:
        # Convert the stack trace to string to send to client
        payload = 'ERR ' + repr(traceback.format_exc())  # repr makes it so multi-line strings go as one string
    else:
        # Tell the client that it worked
        payload = 'ACK ' + str(return_val)
    return payload



''' These are functions specific to phidl.
    There is not necessarily any similarity with implementation in other languages
'''
from __future__ import print_function
import phidl
from phidl import Device

import time
import os
from contextlib import contextmanager
from functools import wraps

# Makes it so that only one import is needed: lyipc.client.phidl will drop in for lyipc.client
from lyipc.client.general import *
from lyipc.client.dependent import *


def trace_phidladd(device, file):
    ''' Writes to file and loads in the remote instance whenever phidl.Device.add is called
    '''
    phidl.device_layout.Device.old_add = phidl.device_layout.Device.add
    def new_add(self, *args, **kwargs):
        retval = phidl.device_layout.Device.old_add(self, *args, **kwargs)
        klayout_quickplot(device, file, fresh=False)
        return retval
    phidl.device_layout.Device.add = new_add



''' This package contains a core for communicating with the server (in this file)
    and an API that is exposed to client-side layout programs. Modules are organized based on package.
        - general.py is package independent
        - pya.py is klayout specific
        - phidl.py is phidl specific
'''
from __future__ import print_function

import lyipc
from lygadgets import isGSI


# Determine which socket class to use
if not isGSI():
    # print('Warning: pya will not be available')
    import sys
    if sys.version_info[0] == 2:
        raise Exception('Please use python >= 3.1 with lyipc for the time being')
    elif sys.version_info[0] == 3:
        import PyQt5.QtNetwork
        from PyQt5.QtNetwork import QTcpSocket
        localhost = PyQt5.QtNetwork.QHostAddress.LocalHost
else:
    from lygadgets import pya
    from pya import QTcpSocket
    localhost = 'localhost'


def send(message='ping 1234', port=lyipc.PORT):
    ''' Sends a raw message

        Trick here is that PyQt5 does not do automatic str<->byte encoding, but pya does. Also something weird with the addresses
    '''
    payload = message + '\r\n'
    psock = QTcpSocket()
    if not isGSI():
        payload = payload.encode()
    psock.connectToHost(localhost, port)
    if psock.waitForConnected():
        psock.write(payload)
        if psock.waitForReadyRead(3000):
            ret = psock.readLine()
            if not isGSI():
                ret = bytes(ret).decode('utf-8')
            handle_query(ret)
        else:
            raise Exception('Not acknowledged')
    else:
        print('Connection Fail! (tried {}:{})'.format(localhost, port))
    # psock.close()


def handle_query(retString):
    ''' Extra check that there is handshaking. Now handles error
    '''
    if retString.startswith('ACK '):
        return retString[4:]
    elif retString.startswith('ERR '):
        traceback_repr = retString[4:]
        # The traceback is sent as a single line as a string repr. Now convert back to multi-line
        traceback_str = traceback_repr.encode("utf-8").decode('unicode_escape').strip("'")
        print('\n' + traceback_str)
        print('^ Server-side error ^')
        sys.exit(1)


from lyipc.client.general import *
from lyipc.client.dependent import *



''' These are functions that are independent of the layout package '''
from __future__ import print_function
from lyipc.client import send
from functools import lru_cache
import os
import time
from functools import wraps


def reload():
    send('reload view')


fast_realpath = lru_cache(maxsize=4)(os.path.realpath)  # Since the argument is going to be the same every time
def load(filename, mode=None):
    '''
        Modes are
        - 0 (default): replacing the current layout view
        - 1: making a new view
        - 2: adding the layout to the current view (mode 2)
    '''
    filename = fast_realpath(filename)
    tokens = ['load', filename]
    if mode is not None:
        tokens.append(str(mode))
    send(' '.join(tokens))


def kill():
    send('kill')


def diff(file1, file2):
    pass  # TODO





''' These are functions specific to pya
    There is not necessarily any similarity with implementation in other languages
'''
from __future__ import print_function
from lygadgets import pya

import time
import os

# Makes it so that only one import is needed: lyipc.client.phidl will drop in for lyipc.client
from lyipc.client.general import *
from lyipc.client.dependent import *


def trace_pyainsert(layout, file):
    ''' Writes to file and loads in the remote instance whenever pya.Shapes.insert is called
        "layout" is what will be written to file and loaded there.

        Intercepts pya.Shapes.insert globally, not just for the argument "layout".
        This is because usually cells are generated before they are inserted into the layout,
        yet we would still like to be able to visualize their creation.
    '''
    pya.Shapes.old_insert = pya.Shapes.insert
    def new_insert(self, *args, **kwargs):
        retval = pya.Shapes.old_insert(self, *args, **kwargs)
        klayout_quickplot(layout, file, fresh=False)
        return retval
    pya.Shapes.insert = new_insert




''' Server is designed to run from klayout GSI, usually in GUI mode

    Current state: only way to stop serving is close the application.
'''
from __future__ import print_function

import socket
import lyipc
from lygadgets import message, isGSI

if not isGSI():
    raise RuntimeError('Non-klayout serving does not make sense')
import pya


# As of now, port is hard coded and there is no way to stop it besided closing the app
# We have to make sure that a second server doesn't come along and clash
global __active_server
__active_server = None
def run_server():
    global __active_server
    if __active_server is None:
        __active_server = KlayoutServer()
    return __active_server


class KlayoutServer(pya.QTcpServer):
    def new_connection(self):
        from lyipc.interpreter import parse_message
        # Handle incoming connection
        connection = self.nextPendingConnection()
        msg = 'null'
        while connection.isOpen() and connection.state() == pya.QTcpSocket.ConnectedState:
            if connection.canReadLine():
                payload = connection.readLine()
                msg = payload.rstrip('\n').rstrip('\r')
                response = parse_message(msg)
                connection.write(response)
                connection.disconnectFromHost()
            else:
                connection.waitForReadyRead(500)
        signal = pya.qt_signal("disconnected()")
        slot = pya.qt_slot("deleteLater()")
        pya.QObject.connect(connection, signal, connection, slot)


    def __init__(self, port=lyipc.PORT, parent=None):
        pya.QTcpServer.__init__(self, parent)
        localhost = pya.QHostAddress()
        self.listen(localhost, port)
        self.newConnection(self.new_connection)
        message('Server initialized with {}, {}'.format(localhost, port))



from setuptools import setup
import os


def readme():
    with open('README.md') as fx:
        return fx.read()


setup(name='lyipc',
      version='0.2.5',
      description='Inter-process communication for Klayout',
      long_description=readme(),
      author='Alex Tait',
      author_email='alexander.tait@nist.gov',
      license='MIT',
      packages=['lyipc', 'lyipc/client'],
      install_requires=['lygadgets>=0.1.7'],
      package_data={'': ['*.lym']},
      include_package_data=True,
      cmdclass={},
      )



''' These depend on the language being used but have very similar implementation across languages.
    For example, every one has some way to write layout objects to a gds file.
'''
from __future__ import print_function
import os
import time
from functools import wraps
from lyipc.client.general import load, reload


global __unbound_method_names
__unbound_method_names = dict()  # this is a cache to avoid lots of importing
def _get_write_method(writable_obj):
    ''' This function resolves write methods for cell/layout objects across languages.

        If the class is a phidl.Device, we will dynamically import gdspy, then check if it is a subclass of gdspy.Cell.
        It is, so then then we return the "write_gds" method which is bound to ``writable_obj``.

        Args:
            (object): layout data object that can be written, typically Cell or Layout or Device

        Returns:
            (func): bound method corresponding to a method that produces a file output (often GDS)
    '''
    global __unbound_method_names

    if type(writable_obj) not in __unbound_method_names.keys():
        # It has not been added to the cache yet. Search for the method in the library.
        try:
            import gdspy  # phidl and others are based on gdspy, so this covers them
            if isinstance(writable_obj, gdspy.Cell):
                __unbound_method_names[type(writable_obj)] = 'write_gds'
        except ImportError:
            pass

        try:
            import pya
            if isinstance(writable_obj, (pya.Cell, pya.Layout)):
                __unbound_method_names[type(writable_obj)] = 'write'
        except ImportError:
            pass

    found_meth_name = __unbound_method_names[type(writable_obj)]
    return getattr(writable_obj, found_meth_name)


def safe_write(writable_obj, filename, write_kwargs=None):
    ''' Writes a temporary file then waits until it finishes writing before moving to the desired filename
    '''
    write_func = _get_write_method(writable_obj)
    dirname, basename = os.path.split(filename)
    temp_filename = os.path.join(dirname, '~.' + basename)
    if write_kwargs is None:
        write_kwargs = dict()
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass
    try:
        write_func(temp_filename, **write_kwargs)
        os.rename(temp_filename, filename)
    except Exception as err:
        try:
            os.remove(temp_filename)
        except FileNotFoundError:
            pass
        raise err


def klayout_quickplot(writable_obj, filename, fresh=False, write_kwargs=None):
    ''' Does the write, wait, and load all in one.

        The write function used is class and language-specific,
        so the ``writable_obj`` is handled according to the _get_write_method function

        Args:
            writable_obj (object): some object that has a resolvable write GDS method (gdspy.Cell, phidl.Device, pya.Cell, pya.Layout)
            filename (str): file to write to
            fresh (bool): determines whether to load (True) or reload (False)
            write_kwargs (dict): kwargs that will be passed to the write function
    '''
    # Write and wait for filename to finish writing
    safe_write(writable_obj, filename, write_kwargs)
    # Tell remote klayout GUI to load/reload it
    if fresh:
        load(filename)
    else:
        reload()


def generate_display_function(default_writable_obj, default_filename):
    ''' A quick way to configure quickplotter into a brief command

        Usage::
            TOP = Layout()
            kqp = make_display_function(TOP, 'debugging.gds')
            ...
            kqp()
    '''
    default_filename = os.path.realpath(default_filename)
    @wraps(klayout_quickplot)
    def k_quick(writable_obj=default_writable_obj, filename=default_filename, **kwargs):
        klayout_quickplot(writable_obj, filename, **kwargs)
    return k_quick